/*
 *  Program to test DC Motors with Quadrature Encoder as Servos
 *  Hardware:
 *  - Arduino Mega
 *  - Motor Shield  L298
 *  - XM Motor with Quadrature output
 *  - Petentiomenter
 *  
 *  This program can:
 *  - drive a Motor full turns back and forth
 *  - drive a Motor angular position in a sinusoidal manner
 *  - control the motor angular position with a potentiometer
 * 
 *  there is no coppy right on this code, use it without guarantee
 *  
 *     by Florian Bauer - Jun 2023
 */
 
 #include <PID_v1.h>
#include <AFMotor.h>
#include <math.h>


#define POTCONTROL 1
#define FULL_TURNS 2
#define SINUS      3

// uncomment one of the folloiwing lines
//#define MODE POTCONTROL  
#define MODE FULL_TURNS
//#define MODE SINUS 



#define SERIAL
 AF_DCMotor mot1(1,MOTOR12_2KHZ);

int encoderPin1 = 18; //Encoder Output 'A' must connected with intreput pin of arduino.
int encoderPin2 = 19; //Encoder Otput 'B' must connected with intreput pin of arduino.
int potPin   =     2;  //Analog input pin for Potentiometer

volatile int lastEncoded = 0;   // Here updated value of encoder store.
volatile long encoderValue = 0; // Raw encoder value

double kp = 250, ki = .01, kd = 0.3;              // modify for optimal performance
double input = 0, output = 0, setpoint = 0;
int limit = 222;

PID myPID  =   PID (&input, &output, &setpoint, kp, ki, kd, DIRECT);  // PID-Controller for Servo
unsigned long myTime;
unsigned long startTime;
int User_Input0=0;

void setup() {
  Serial.begin(250000); //initialize serial comunication

  pinMode(encoderPin1, INPUT_PULLUP); 
  digitalWrite(encoderPin1, HIGH); //turn pullup resistor on
  attachInterrupt(digitalPinToInterrupt(encoderPin1), updateEncoder, CHANGE);

  pinMode(encoderPin2, INPUT_PULLUP);
  digitalWrite(encoderPin2, HIGH); //turn pullup resistor on
  attachInterrupt(digitalPinToInterrupt(encoderPin2), updateEncoder, CHANGE);

  myPID.SetMode(AUTOMATIC);   //set PID in Auto mode
  myPID.SetSampleTime(1);  // refresh rate of PID controller
  myPID.SetOutputLimits(-limit, limit); // this is the MAX PWM value to move motor, here change in value reflect change in speed of motor.
  User_Input0 = analogRead(2); // liest das Potentiometer aus (Wert zwischen 0 und 1023)


 startTime = millis();
}

int setpoints [] = {0,253,0, -253 };

int is=0;
long oldTime=0;
int scharf=1;
void loop() {
  
  int User_Input = analogRead(potPin); // liest das Potentiometer aus (Wert zwischen 0 und 1023)
  myTime = millis();
   if( abs(input-setpoint) < 2)
  {
    if(scharf)
    { Serial.println(myTime-oldTime);
    scharf=0;
    }
  }
#if MODE == POTCONTROL  
  setpoint = map(User_Input, 0, 1023, -253, 253);
#endif

#if MODE == FULL_TURNS
  if(myTime-oldTime >= 2000)
  {
    setpoint=setpoints[is];
    is = (is+1) % 4;
    oldTime=myTime;
    scharf=1;
  }
#endif
#if MODE == SINUS 
  setpoint = (int) (260 * sin( (myTime-oldTime)*0.0001*10.0*PI));
#endif  
  
  input = encoderValue; // data from encoder consider as a Process value
 
  myPID.Compute();                 // calculate new output
  pwmOut(0,output);
 
  Serial.print("T:");    Serial.print(myTime-startTime);
  Serial.print(" SOLL:");    Serial.print(setpoint);
  Serial.print(" IST:");    Serial.print(input);
  Serial.print(" STELL:");  Serial.println(output);
}

void pwmOut(int i,int out) {                  
  if (out > 0) {                  
    mot1.run(FORWARD); mot1.setSpeed(out); 
  }
  else {
    mot1.run(BACKWARD); mot1.setSpeed(abs(out)); 
  }
}

void updateEncoder(){
  int MSB = digitalRead(encoderPin1); //MSB = most significant bit
  int LSB = digitalRead(encoderPin2); //LSB = least significant bit

  int encoded = (MSB << 1) |LSB; //converting the 2 pin value to single number
  int sum  = (lastEncoded << 2) | encoded; //adding it to the previous encoded value

  if(sum == 0b1101 || sum == 0b0100 || sum == 0b0010 || sum == 0b1011) encoderValue ++;
  if(sum == 0b1110 || sum == 0b0111 || sum == 0b0001 || sum == 0b1000) encoderValue --;

  lastEncoded = encoded; //store this value for next time
}
